"""
CUSTOM READERS MODULE FOR GOOGLE POSTMASTER READER
"""
from typing import Dict, List, Any, Union, Generator
from googleapiclient import discovery, errors
from google.oauth2 import service_account

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.date_managers import (
    date_range_iterator,
    check_date_range_validity,
)
from sdc_dp_helpers.base_readers import BaseReader


class GooglePostmasterReader(BaseReader):
    """
    Google Postmaster Reader
    """

    def __init__(self, creds_filepath: str, config_filepath: str) -> None:
        super().__init__()
        self.secrets: Dict[Any, Any] = load_file(creds_filepath)
        self.config: Dict[Any, Any] = load_file(config_filepath)
        self.service = self._get_auth()
        self.success: List[bool] = []

    def _get_auth(self):
        """
        Get our credentials initialised above and use those to get client.
        """
        credentials = service_account.Credentials.from_service_account_info(
            info=self.secrets,
            scopes=["https://www.googleapis.com/auth/postmaster.readonly"],
        )
        service = discovery.build(
            serviceName="gmailpostmastertools",
            version="v1beta1",
            credentials=credentials,
        )
        if service is None:
            raise RuntimeError("Service is null.")
        if not hasattr(service, "domains"):
            raise KeyError("Service does not have domains attribute.")
        return service

    def _query_handler(self, *args, **kwargs) -> dict:
        # pylint: disable=no-member
        """Handles the Query call"""

        if not {"domain", "date"}.issubset(set(kwargs.keys())):
            raise KeyError("Invalid arguments - expecting: domain, date")
        response = (
            self.service.domains()
            .trafficStats()
            .get(name=f"domains/{kwargs['domain']}/trafficStats/{kwargs['date']}")
            .execute()
        )
        if response is None:
            raise RuntimeError("response is 'None'")
        self.is_success()
        return response

    def run_query(
        self,
    ) -> Union[
        Generator[Dict[List[Dict[Any, Any]], Any], None, None],
        Dict[List[Dict[Any, Any]], Any],
    ]:
        """Run the query and return the data"""
        check_date_range_validity(self.config["start_date"], self.config["end_date"])

        domains = self.config["domains"]

        date_range = date_range_iterator(
                start_date=self.config["start_date"],
                end_date=self.config["end_date"],
                interval="1_day",
                end_inclusive=True,
                time_format="%Y%m%d",
            )

        for domain in domains:
            print(f"domain : {domain}")
            for date, _ in date_range:
                print(f"date : {date}")
                try:

                    dataset = self._query_handler(domain=domain, date=date)
                    print(f"{domain} {date} success")
                    yield {
                        "date": date,
                        "brand": domain.replace(".", "_"),
                        "data": [dataset],
                    }
                except RuntimeError as err:
                    self.not_success()
                    print(f"Error occured: {err}")
                except errors.HttpError as err:
                    self.not_success()
                    print(f"{domain} {date} fail: {err}")
                except KeyError as err:
                    self.not_success()
                    print(f"An error occurred: {err}")



"""
    CUSTOM READER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
# pylint: disable=no-member,too-many-locals,too-many-arguments,broad-except,too-few-public-methods,arguments-differ,line-too-long,broad-exception-raised,inconsistent-return-statements
from datetime import datetime

from google.oauth2 import service_account
from google.cloud import enterpriseknowledgegraph as ekg
from sdc_dp_helpers.google_knowledge_graph.gkg_sdk import EnityIdsEndpoints, QueriesEndpoints

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.data_managers import multiple_regex_replace
from sdc_dp_helpers.base_readers import BaseReader


class GoogleKnowledgeGraphReader(BaseReader):
    """Google Knowledge Graph Class"""

    def __init__(self, secrets_filepath: str, config_filepath: str):
        self.secrets: str = secrets_filepath
        self.config: dict = load_file(config_filepath)
        self.client = self._get_auth()
        self.dataset: list = []
        self.success = []
        self.api_calls = 0

    def _get_auth(self):
        """
        Get our credentials initialised above and use those to get client

        """
        credentials = service_account.Credentials.from_service_account_file(
            self.secrets
        )
        client = ekg.EnterpriseKnowledgeGraphServiceClient(credentials=credentials)
        return client

    def format_payload(self, payload, query, language, date, entity_id):
        if payload:
            query_partition: str = self.query_partition(query, language)
            payload.update(
                {
                    "date": date,
                    "query": query,
                    "language": language,
                    "query_partition": query_partition,
                    "entity_id": entity_id,
                }
            )
            self.is_success()
            print(f"Got data for query:'{query}' language:'{language}' on '{date}'")
        if not payload:
            self.not_success()
            print(f"No data for query:'{query}' language:'{language}' on '{date}'")
        return payload

    def get_data(self, endpoints, date, results):
        if endpoints[0] == "entity_ids":
            for query, entity_language in endpoints[1].items():
                languages = list(entity_language[0].values())[0]
                entity_id = list(entity_language[0].keys())[0]
                for language in languages:
                    payload = EnityIdsEndpoints(
                        client=self.client, config=self.config
                    ).make_api_call(entity_id, language)
                    payload = self.format_payload(
                        payload, query, language, date, entity_id
                    )
                    if payload:
                        results.append(payload)
        if endpoints[0] == "queries":
            for query, languages in endpoints[1].items():
                for language in languages:
                    entity_id = None
                    payload = QueriesEndpoints(
                        client=self.client, config=self.config
                    ).make_api_call(query, language)
                    payload = self.format_payload(
                        payload, query, language, date, entity_id
                    )
                    if payload:
                        results.append(payload)
        return results

    def query_partition(self, query: str, language: str = None) -> str:
        """Helper to get the query partition string
        :query: "string",
        :language:"string"}
        :returns: query_partition - str
        """
        query = query.strip()
        query_partition = multiple_regex_replace(
            query, {".": "_dot_", " ": "_space_", "+": "_plus_"}
        )
        if language:
            query_partition = f"{query_partition}_lang_{language}"

        return query_partition

    def _query_handler(self):
        results=[]
        date = datetime.strftime(datetime.now(), "%Y-%m-%d")
        for endpoints in self.config["endpoints"].items():
            results = self.get_data(endpoints, date, results)
        if results:
            return {"date": date, "data": results}
        self.dataset = results

    def run_query(self):
        output = self._query_handler()
        return output



# pylint: disable=too-few-public-methods, unused-import ,attribute-defined-outside-init,broad-exception-raise,trailing-whitespace,line-too-long,no-member,broad-exception-raised,bare-except,too-many-arguments,arguments-differ,wildcard-import,broad-exception-caught,unused-wildcard-import
import os
import json
import ast
from abc import abstractmethod
import requests
from oauth2 import *
from xero.auth import OAuth2Credentials
from xero.exceptions import XeroForbidden
from xero import Xero
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler


class XeroQuotaException(Exception):
    """Class for Xero  Quota Exception"""


class Authenticator:
    """
    Xero authentication and token refresh
    """

    def __init__(self, config, creds_filepath):
        self.creds_path: str = creds_filepath
        self.config: dict = config

    def token_isvalid(self, creds: OAuth2Credentials) -> bool:
        if not isinstance(creds, OAuth2Credentials):
            raise TypeError("creds is not an object of type OAuth2Credentials")

        return creds.get_tenants()[0]["tenantId"] is not None

    def get_auth_token(self) -> OAuth2Credentials:
        """
        Consumes the client id and the previous auth processes refresh token.
        This returns an authentication token that will last 30 minutes
        to make queries the minute it is used. Or it will expire in 60 days of no use.
        The newly generated last refresh token now needs token stored for
        next use.
        PS: we receive and save the auth_token in a local dir supplied by the encapsulating project
        we never interact with s3 from inside of this class
        """
        self.client_id = self.config.get("client_id", None)
        if not self.client_id:
            raise ValueError("No client_id set")
        if (not self.client_id) or (len(self.client_id) != 32):
            raise ValueError("Invalid client_id")

        with open(self.creds_path, "r") as token_file:
            auth_token = ast.literal_eval(token_file.read())

        auth_creds = OAuth2Credentials(
            self.client_id, client_secret="", token=auth_token
        )
        return self.refresh_auth_token(auth_creds)

    def refresh_auth_token(self, auth_creds: OAuth2Credentials) -> OAuth2Credentials:
        cred = {
            "grant_type": "refresh_token",
            "refresh_token": auth_creds.token["refresh_token"],  #
            "client_id": self.config.get("client_id", None),
        }
        response = requests.post(
            "https://identity.xero.com/connect/token", cred, timeout=30
        )
        auth_token = response.json()
        err_message = auth_token.get("error")
        if err_message:
            raise Exception(err_message)
        with open(self.creds_path, "w") as outfile:
            outfile.write(json.dumps(auth_token))

        auth_creds = OAuth2Credentials(
            self.client_id, client_secret="", token=auth_token
        )
        if not self.token_isvalid(auth_creds):
            raise XeroForbidden(
                f"Error while trying to authenticate the refreshed token: {str(auth_creds)}"
            )
        return auth_creds


class RequestHandler:
    """Interface for  API Call method"""

    def __init__(self, config, creds_filepath):
        self.creds_path: str = creds_filepath
        self.config: dict = config
        self.authenticator = Authenticator(
            config=self.config, creds_filepath=self.creds_path
        )

    @abstractmethod
    def make_api_call(self, **kwargs):
        """Make API Call"""
        raise NotImplementedError


class XeroAPICall:
    "Class for making xero API call"

    def __init__(self, creds_path, config):
        self.creds_filepath = creds_path
        self.config = config

    def get_reports(self, report_name):
        results = []
        get_data_methods = {
            "filtered": ReportsByTrackingOption,
            "notfiltered": ReportsNotFiltered,
        }
        for filterby in self.config["reports"][report_name]["filter"]:
            method_ids_caller = get_data_methods[filterby](
                config=self.config, creds_filepath=self.creds_filepath
            )
            data = method_ids_caller.make_api_call(
                report_name, self.config["reports"][report_name]["params"], filterby
            )
            results.append(data)
        return results

    def get_modules(self, module):
        module_ = Modules(config=self.config, creds_filepath=self.creds_filepath)
        results = module_.make_api_call(module)
        return results


class ReportsNotFiltered(RequestHandler):
    "Class for not Filtered reports"
    def make_api_call(self, report_name, request_params, filterby):
        data_set = {}
        auth_token = self.authenticator.get_auth_token()
        trackingcategories = GetTrackingCategories(
            config=self.config, creds_filepath=self.creds_path
        ).make_api_call()
        params = {}
        option_id = None
        for tracking_category in trackingcategories:
            file_name = (
                report_name
                + "_"
                + filterby
                + "_"
                + tracking_category["TrackingCategoryID"]
            )
            params.update(
                {"trackingCategoryID": tracking_category["TrackingCategoryID"]}
            )
            report = get_report(
                auth_token,
                params,
                option_id,
                self.config["tenant_id"],
                report_name,
                request_params,
            )
            result = {
                "endpoint": report_name,
                "tenant_name": self.config["tenant_name"],
                "trackingCategoryId": tracking_category["TrackingCategoryID"],
                "data": [report],
                "date": request_params.get("date", request_params.get("end_date")),
                "filterby": filterby,
            }
            if result and result is not None:
                data_set.update({file_name: result})
        return data_set


class GetTrackingCategories(RequestHandler):
    "class to get Reports filtered by Tracking categories"

    def make_api_call(self):
        auth_token = self.authenticator.get_auth_token()
        auth_token.tenant_id = self.config.get("tenant_id")
        xero_obj = Xero(auth_token)
        trackingcategories = (
            i for i in xero_obj.trackingcategories.all() if i is not None
        )
        return trackingcategories


class ReportsByTrackingOption(RequestHandler):
    "class to get Reports filtered by TrackingOptionID"

    def make_api_call(self, report_name, request_params, filterby):
        data_set = {}
        trackingcategories = GetTrackingCategories(
            config=self.config, creds_filepath=self.creds_path
        ).make_api_call()
        auth_token = self.authenticator.get_auth_token()
        params = {}

        for tracking_category in trackingcategories:
            option_data = []
            for option_id in tracking_category["Options"]:
                file_name = (
                    report_name
                    + "_"
                    + filterby
                    + "_"
                    + tracking_category["TrackingCategoryID"]
                )
                params.update(
                    {"trackingCategoryID": tracking_category["TrackingCategoryID"]}
                )
                params.update({"TrackingOptionID": option_id["TrackingOptionID"]})
                report = get_report(
                    auth_token,
                    params,
                    option_id,
                    self.config["tenant_id"],
                    report_name,
                    request_params,
                )
                option_data.append(report)
            result = {
                "endpoint": report_name,
                "tenant_name": self.config["tenant_name"],
                "trackingCategoryId": tracking_category["TrackingCategoryID"],
                "data": option_data,
                "date": request_params.get("date", request_params.get("end_date")),
                "filterby": filterby,
            }
            if option_data and option_data is not None:
                data_set.update({file_name: result})
        return data_set


class Modules(RequestHandler):
    "class to get Modules"

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def run_request(self, xero_client, api_object, request):
        """
        Run the API request that consumes a request payload and site url.
        This separates the request with the request handler from the rest of the logic.
        """
        # To Do Handle API Errors
        api_call = getattr(xero_client, api_object)
        # XeroRateLimitExceeded
        return api_call.filter(
            raw=date_filter_helper(request["from_date"], request["to_date"]),
            page=request["page"],
        )

    def make_api_call(self, module):
        auth_token = self.authenticator.get_auth_token()
        auth_token.tenant_id = self.config.get("tenant_id")
        xero = Xero(auth_token)
        data_set = {}
        if module not in [
            "accounts",
        ]:
            raise ValueError(module + " is not supported or does not exist.")
        data_set[module] = []
        prev_response = None
        page = 1
        while True:
            response = self.run_request(
                xero_client=xero,
                api_object=module,
                request={
                    "from_date": self.config.get("start_date"),
                    "to_date": self.config.get("end_date"),
                    "page": page,
                },
            )
            print(response)
            if len(response) < 1:
                print("Request returned empty payload. breaking...")
                break
            if response == prev_response:
                print("Request returned copy of last payload. breaking...")
            data_set[module] += [
                json.loads(
                    json.dumps(response_obj, indent=4, sort_keys=True, default=str)
                )
                for response_obj in response
            ]
            # ensure the token is still fresh
            auth_token = self.authenticator.get_auth_token()
            prev_response = response
            page += 1
        return data_set


############### helper functions



@retry_handler(exceptions=XeroQuotaException, total_tries=5, initial_wait=5)
def get_report(auth_token, params, option_id, tenant_id, report_name, request_params):
    if report_name in ["ProfitAndLoss"]:
        params = {
            "fromDate": request_params["start_date"],
            "toDate": request_params["end_date"],
        }
    if report_name in ["BalanceSheet"]:
        params.update({"date": request_params.get("date")})
        params.update({"timeframe": request_params.get("timeframe")})
        params.update({"periods": request_params.get("periods")})
    headers = {
        "Authorization": "Bearer " + auth_token.token["access_token"],
        "Xero-Tenant-Id": tenant_id,
        "Accept": "application/json",
    }
    try:
        response = requests.get(
            "https://api.xero.com/api.xro/2.0/Reports/" + report_name,
            params=params,
            headers=headers,
            timeout=30,
        )
        report = json.loads(
            response.text.replace("\r", "").replace("\n", "").strip("'<>() ")
        )
        report = format_values(report, tenant_id, option_id=option_id)
    except Exception as err:
        if err.code == 429:
            raise XeroQuotaException(
                f"Xero Quota Reached" f"Status code: {err.code}, Reason: {err.reason}. "
            ) from err
    return report


def format_values(report, tenant_id, option_id=None):
    for row in range(len(report["Reports"][0]["Rows"])):
        period = report["Reports"][0]["Rows"][0]["Cells"]
        try:
            for report_ in report["Reports"][0]["Rows"][row]["Rows"]:
                for idx, cells in enumerate(report_["Cells"]):
                    cells["Attributes"][0].update({"type": report_["Cells"][0]["Value"]})
                    cells["Attributes"][0].update({"period": period[idx]["Value"]})
        except:
            pass
    if option_id:
        report["TrackingOptionID"] = option_id["TrackingOptionID"]
        report["TrackingOptionName"] = option_id["Name"]
    report["tenantId"] = tenant_id
    return report


def date_filter_helper(from_date: str, to_date: str, filter_field: str = None) -> str:
    """
    Custom implementation of date filters borrowed from:
    https://github.com/ClaimerApp/pyxero/blob/master/xero/basemanager.py
    """
    if not from_date:
        raise ValueError("No from_date set")

    # common date_field inside of the accounts and contacts modules is UpdatedDateUTC
    filter_field = "UpdatedDateUTC" if not filter_field else filter_field
    api_filter = filter_field + ">=DateTime(" + ",".join(from_date.split("-")) + ")"
    if to_date:
        api_filter = (
            api_filter
            + "&&"
            + filter_field
            + "<=DateTime("
            + ",".join(to_date.split("-"))
            + ")"
        )
    # end if
    return api_filter



# pylint: disable=too-few-public-methods,super-init-not-called,no-member,broad-exception-raised,attribute-defined-outside-init,arguments-differ,import-error,too-many-arguments,too-many-locals,line-too-long, disable=useless-elif-before-else, disable=R1705, disable=R1720, disable=W1309, disable=C0103
""" GKG Reader SDK"""

import os
from google.cloud import enterpriseknowledgegraph as ekg
from google.protobuf.json_format import MessageToDict
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler

class ServiceUnavailableException(Exception):
    """Class for GKG V2  Service Unavailable Exception"""


class GoogleKnowledgeGraphQuotaException(Exception):
    """Class for GKG V2  Quota Exception"""


class RequestHandler:
    """Interface for API call"""

    api_calls = 0

    def __init__(self, creds):
        self._creds = creds

    def make_api_call(self, **kwargs):
        """Make API call"""
        raise NotImplementedError


@request_handler(
        wait=float(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method='random'
    )
@retry_handler(
    exceptions=GoogleKnowledgeGraphQuotaException,
    total_tries=5,
    initial_wait=61,
    backoff_factor=1.2,
)
@retry_handler(exceptions=ServiceUnavailableException, total_tries=5, initial_wait=100)
class QueriesEndpoints(RequestHandler):
    """ GKG QueriesEndpoints SDK"""
    def __init__(self, client, config):
        self.client = client
        self.config = config

    def make_api_call(self, query: str, language: str):
        RequestHandler.api_calls += 1
        response_json = {}
        parent = self.client.common_location_path(
            project=self.config.get("project_id"), location=self.config.get("location")
        )
        request = ekg.SearchRequest(
            parent=parent,
            query=query,
            languages=[language],
            types=self.config.get("types"),
            limit=self.config.get("limit", 1),
        )
        try:
            response = self.client.search(request=request)
            response_json = MessageToDict(
                response._pb, including_default_value_fields=True
            )
        except Exception as err:
            if err.code == 429:
                raise GoogleKnowledgeGraphQuotaException(
                    f"GoogleKnowledgeGraph Quota Reached"
                    f"Status code: {err.code}, Reason: {err.reason}. "
                ) from err
            if err.code in [404, 500, 202]:
                raise ConnectionError(
                    f"Google Knowledge Graph API failed. "
                    f"Status code: {err.code}, Reason: {err.reason}."
                ) from err
            if err.code in [503]:
                raise ServiceUnavailableException(
                    f"Google Knowledge Graph API failed. "
                    f"Status code: {err.code}, Reason: {err.reason}."
                ) from err
            raise Exception(f"Unexpected error: {err}") from err
        return response_json


@request_handler(
        wait=float(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method='random'
    )
@retry_handler(
    exceptions=GoogleKnowledgeGraphQuotaException,
    total_tries=5,
    initial_wait=61,
    backoff_factor=1.2,
)
@retry_handler(exceptions=ServiceUnavailableException, total_tries=5, initial_wait=100)
class EnityIdsEndpoints(RequestHandler):
    """ GKG EnityIdsEndpoints SDK"""
    def __init__(self, client, config):
        self.client = client
        self.config = config

    def make_api_call(self, entity_id: str, language: str):
        RequestHandler.api_calls += 1
        response_json = {}
        parent = self.client.common_location_path(
            project=self.config.get("project_id"), location=self.config.get("location")
        )
        request = ekg.LookupRequest(
            parent=parent, ids=[entity_id], languages=[language]
        )
        try:
            response = self.client.lookup(request=request)
            response_json = MessageToDict(
                response._pb, including_default_value_fields=True
            )
        except Exception as err:
            if err.code == 429:
                raise GoogleKnowledgeGraphQuotaException(
                    f"GoogleKnowledgeGraph Quota Reached"
                    f"Status code: {err.code}, Reason: {err.reason}. "
                ) from err
            if err.code in [404, 500, 202]:
                raise ConnectionError(
                    f"Google Knowledge Graph API failed. "
                    f"Status code: {err.code}, Reason: {err.reason}."
                ) from err
            if err.code in [503]:
                raise ServiceUnavailableException(
                    f"Google Knowledge Graph API failed. "
                    f"Status code: {err.code}, Reason: {err.reason}."
                ) from err
            raise Exception(f"Unexpected error: {err}") from err
        return response_json
